/* ANSI-C code produced by gperf version 3.3 */
/* Command-line: gperf --output-file src/udev/link-config-gperf.c ../systemd/src/udev/net/link-config-gperf.gperf  */
/* Computed positions: -k'6,12,17-18,20,$' */

#if !((' ' == 32) && ('!' == 33) && ('"' == 34) && ('#' == 35) \
      && ('%' == 37) && ('&' == 38) && ('\'' == 39) && ('(' == 40) \
      && (')' == 41) && ('*' == 42) && ('+' == 43) && (',' == 44) \
      && ('-' == 45) && ('.' == 46) && ('/' == 47) && ('0' == 48) \
      && ('1' == 49) && ('2' == 50) && ('3' == 51) && ('4' == 52) \
      && ('5' == 53) && ('6' == 54) && ('7' == 55) && ('8' == 56) \
      && ('9' == 57) && (':' == 58) && (';' == 59) && ('<' == 60) \
      && ('=' == 61) && ('>' == 62) && ('?' == 63) && ('A' == 65) \
      && ('B' == 66) && ('C' == 67) && ('D' == 68) && ('E' == 69) \
      && ('F' == 70) && ('G' == 71) && ('H' == 72) && ('I' == 73) \
      && ('J' == 74) && ('K' == 75) && ('L' == 76) && ('M' == 77) \
      && ('N' == 78) && ('O' == 79) && ('P' == 80) && ('Q' == 81) \
      && ('R' == 82) && ('S' == 83) && ('T' == 84) && ('U' == 85) \
      && ('V' == 86) && ('W' == 87) && ('X' == 88) && ('Y' == 89) \
      && ('Z' == 90) && ('[' == 91) && ('\\' == 92) && (']' == 93) \
      && ('^' == 94) && ('_' == 95) && ('a' == 97) && ('b' == 98) \
      && ('c' == 99) && ('d' == 100) && ('e' == 101) && ('f' == 102) \
      && ('g' == 103) && ('h' == 104) && ('i' == 105) && ('j' == 106) \
      && ('k' == 107) && ('l' == 108) && ('m' == 109) && ('n' == 110) \
      && ('o' == 111) && ('p' == 112) && ('q' == 113) && ('r' == 114) \
      && ('s' == 115) && ('t' == 116) && ('u' == 117) && ('v' == 118) \
      && ('w' == 119) && ('x' == 120) && ('y' == 121) && ('z' == 122) \
      && ('{' == 123) && ('|' == 124) && ('}' == 125) && ('~' == 126))
/* The character set is not based on ISO-646.  */
#error "gperf generated tables don't work with this execution character set. Please report a bug to <bug-gperf@gnu.org>."
#endif

#line 2 "../systemd/src/udev/net/link-config-gperf.gperf"

_Pragma("GCC diagnostic ignored \"-Wimplicit-fallthrough\"")
#if __GNUC__ >= 15
_Pragma("GCC diagnostic ignored \"-Wzero-as-null-pointer-constant\"")
#endif
#include <stddef.h>
#include "condition.h"
#include "conf-parser.h"
#include "ethtool-util.h"
#include "link-config.h"
#include "net-condition.h"
#include "netif-sriov.h"
#include "socket-util.h"
#include <string.h>

#define TOTAL_KEYWORDS 128
#define MIN_WORD_LENGTH 8
#define MAX_WORD_LENGTH 41
#define MIN_HASH_VALUE 10
#define MAX_HASH_VALUE 236
/* maximum key range = 227, duplicates = 0 */

#ifdef __GNUC__
__inline
#else
#ifdef __cplusplus
inline
#endif
#endif
static unsigned int
link_config_gperf_hash (register const char *str, register size_t len)
{
  static const unsigned char asso_values[] =
    {
      237, 237, 237, 237, 237, 237, 237, 237, 237, 237,
      237, 237, 237, 237, 237, 237, 237, 237, 237, 237,
      237, 237, 237, 237, 237, 237, 237, 237, 237, 237,
      237, 237,   5, 237, 237, 237, 237, 237, 237, 237,
      237, 237,  60, 237, 237,  55,   0,  25, 237, 237,
      237, 237, 237, 237, 237, 237, 237, 237, 237, 237,
      237, 237, 237, 237, 237,  15,   0,   5,  15,  35,
       60,  30,   0,  25, 237, 237,  85, 120,   0,   5,
       90,   0,   0,   0,   5,  45,  10,  40, 237, 237,
      237, 237, 237, 237, 237, 237, 237,   5, 237,  10,
       15,   5,   0,   5,  65,   0, 237,  30,  30,  85,
       20,  25,   0,   0,   0,   5,   0,  50,  35,  70,
       65,  15,   0, 237, 237, 237, 237, 237, 237, 237,
      237, 237, 237, 237, 237, 237, 237, 237, 237, 237,
      237, 237, 237, 237, 237, 237, 237, 237, 237, 237,
      237, 237, 237, 237, 237, 237, 237, 237, 237, 237,
      237, 237, 237, 237, 237, 237, 237, 237, 237, 237,
      237, 237, 237, 237, 237, 237, 237, 237, 237, 237,
      237, 237, 237, 237, 237, 237, 237, 237, 237, 237,
      237, 237, 237, 237, 237, 237, 237, 237, 237, 237,
      237, 237, 237, 237, 237, 237, 237, 237, 237, 237,
      237, 237, 237, 237, 237, 237, 237, 237, 237, 237,
      237, 237, 237, 237, 237, 237, 237, 237, 237, 237,
      237, 237, 237, 237, 237, 237, 237, 237, 237, 237,
      237, 237, 237, 237, 237, 237, 237, 237, 237, 237,
      237, 237, 237, 237, 237, 237
    };
  register unsigned int hval = len;

  switch (hval)
    {
      default:
        hval += asso_values[(unsigned char)str[19]];
#if (defined __cplusplus && (__cplusplus >= 201703L || (__cplusplus >= 201103L && defined __clang__ && __clang_major__ + (__clang_minor__ >= 9) > 3))) || (defined __STDC_VERSION__ && __STDC_VERSION__ >= 202000L && ((defined __GNUC__ && __GNUC__ >= 10) || (defined __clang__ && __clang_major__ >= 9)))
      [[fallthrough]];
#elif (defined __GNUC__ && __GNUC__ >= 7) || (defined __clang__ && __clang_major__ >= 10)
      __attribute__ ((__fallthrough__));
#endif
      /*FALLTHROUGH*/
      case 19:
      case 18:
        hval += asso_values[(unsigned char)str[17]];
#if (defined __cplusplus && (__cplusplus >= 201703L || (__cplusplus >= 201103L && defined __clang__ && __clang_major__ + (__clang_minor__ >= 9) > 3))) || (defined __STDC_VERSION__ && __STDC_VERSION__ >= 202000L && ((defined __GNUC__ && __GNUC__ >= 10) || (defined __clang__ && __clang_major__ >= 9)))
      [[fallthrough]];
#elif (defined __GNUC__ && __GNUC__ >= 7) || (defined __clang__ && __clang_major__ >= 10)
      __attribute__ ((__fallthrough__));
#endif
      /*FALLTHROUGH*/
      case 17:
        hval += asso_values[(unsigned char)str[16]];
#if (defined __cplusplus && (__cplusplus >= 201703L || (__cplusplus >= 201103L && defined __clang__ && __clang_major__ + (__clang_minor__ >= 9) > 3))) || (defined __STDC_VERSION__ && __STDC_VERSION__ >= 202000L && ((defined __GNUC__ && __GNUC__ >= 10) || (defined __clang__ && __clang_major__ >= 9)))
      [[fallthrough]];
#elif (defined __GNUC__ && __GNUC__ >= 7) || (defined __clang__ && __clang_major__ >= 10)
      __attribute__ ((__fallthrough__));
#endif
      /*FALLTHROUGH*/
      case 16:
      case 15:
      case 14:
      case 13:
      case 12:
        hval += asso_values[(unsigned char)str[11]];
#if (defined __cplusplus && (__cplusplus >= 201703L || (__cplusplus >= 201103L && defined __clang__ && __clang_major__ + (__clang_minor__ >= 9) > 3))) || (defined __STDC_VERSION__ && __STDC_VERSION__ >= 202000L && ((defined __GNUC__ && __GNUC__ >= 10) || (defined __clang__ && __clang_major__ >= 9)))
      [[fallthrough]];
#elif (defined __GNUC__ && __GNUC__ >= 7) || (defined __clang__ && __clang_major__ >= 10)
      __attribute__ ((__fallthrough__));
#endif
      /*FALLTHROUGH*/
      case 11:
      case 10:
      case 9:
      case 8:
      case 7:
      case 6:
        hval += asso_values[(unsigned char)str[5]];
        break;
    }
  return hval + asso_values[(unsigned char)str[len - 1]];
}

const struct ConfigPerfItem *
link_config_gperf_lookup (register const char *str, register size_t len)
{
#if (defined __GNUC__ && __GNUC__ + (__GNUC_MINOR__ >= 6) > 4) || (defined __clang__ && __clang_major__ >= 3)
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wmissing-field-initializers"
#endif
  static const struct ConfigPerfItem wordlist[] =
    {
      {(char*)0}, {(char*)0}, {(char*)0}, {(char*)0},
      {(char*)0}, {(char*)0}, {(char*)0}, {(char*)0},
      {(char*)0}, {(char*)0},
#line 35 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Match.Host",                                config_parse_net_condition,            CONDITION_HOST,                offsetof(LinkConfig, conditions)},
      {(char*)0},
#line 31 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Match.Driver",                              config_parse_match_strv,               0,                             offsetof(LinkConfig, match.driver)},
      {(char*)0},
#line 52 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.Name",                                 config_parse_ifname,                   0,                             offsetof(LinkConfig, name)},
#line 32 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Match.Type",                                config_parse_match_strv,               0,                             offsetof(LinkConfig, match.iftype)},
      {(char*)0}, {(char*)0}, {(char*)0}, {(char*)0},
#line 94 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.ReceiveFCS",                           config_parse_tristate,                 0,                             offsetof(LinkConfig, features[NET_DEV_FEAT_RXFCS])},
      {(char*)0},
#line 147 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"SR-IOV.Trust",                              config_parse_sr_iov_boolean,           0,                             offsetof(LinkConfig, sr_iov_by_section)},
#line 75 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.ScatterGather",                        config_parse_tristate,                 0,                             offsetof(LinkConfig, features[NET_DEV_FEAT_SG])},
#line 42 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Match.Firmware",                            config_parse_net_condition,            CONDITION_FIRMWARE,            offsetof(LinkConfig, conditions)},
#line 33 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Match.Kind",                                config_parse_match_strv,               0,                             offsetof(LinkConfig, match.kind)},
#line 103 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.RxMiniBufferSize",                     config_parse_ring_buffer_or_channel,   0,                             offsetof(LinkConfig, ring.rx_mini)},
      {(char*)0},
#line 41 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Match.Architecture",                        config_parse_net_condition,            CONDITION_ARCHITECTURE,        offsetof(LinkConfig, conditions)},
#line 34 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Match.Property",                            config_parse_match_property,           0,                             offsetof(LinkConfig, match.property)},
#line 55 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.Alias",                                config_parse_ifalias,                  0,                             offsetof(LinkConfig, alias)},
#line 148 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"SR-IOV.LinkState",                          config_parse_sr_iov_link_state,        0,                             offsetof(LinkConfig, sr_iov_by_section)},
#line 102 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.RxBufferSize",                         config_parse_ring_buffer_or_channel,   0,                             offsetof(LinkConfig, ring.rx)},
      {(char*)0},
#line 66 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.Advertise",                            config_parse_advertise,                0,                             offsetof(LinkConfig, advertise)},
#line 76 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.ScatterGatherFragmentList",            config_parse_tristate,                 0,                             offsetof(LinkConfig, features[NET_DEV_FEAT_FRAGLIST])},
#line 27 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Match.MACAddress",                          config_parse_hw_addrs,                 0,                             offsetof(LinkConfig, match.hw_addr)},
#line 105 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.TxBufferSize",                         config_parse_ring_buffer_or_channel,   0,                             offsetof(LinkConfig, ring.tx)},
#line 57 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.ReceiveQueues",                        config_parse_rx_tx_queues,             0,                             offsetof(LinkConfig, rxqueues)},
      {(char*)0},
#line 97 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.RxChannels",                           config_parse_ring_buffer_or_channel,   0,                             offsetof(LinkConfig, channels.rx)},
#line 113 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.RxCoalesceIrqSec",                     config_parse_coalesce_sec,             0,                             offsetof(LinkConfig, coalesce.rx_coalesce_usecs_irq)},
#line 128 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.RxCoalesceHighSec",                    config_parse_coalesce_sec,             0,                             offsetof(LinkConfig, coalesce.rx_coalesce_usecs_high)},
      {(char*)0}, {(char*)0},
#line 98 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.TxChannels",                           config_parse_ring_buffer_or_channel,   0,                             offsetof(LinkConfig, channels.tx)},
#line 117 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.TxCoalesceIrqSec",                     config_parse_coalesce_sec,             0,                             offsetof(LinkConfig, coalesce.tx_coalesce_usecs_irq)},
#line 130 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.TxCoalesceHighSec",                    config_parse_coalesce_sec,             0,                             offsetof(LinkConfig, coalesce.tx_coalesce_usecs_high)},
#line 111 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.RxCoalesceSec",                        config_parse_coalesce_sec,             0,                             offsetof(LinkConfig, coalesce.rx_coalesce_usecs)},
      {(char*)0},
#line 95 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.ReceiveAll",                           config_parse_tristate,                 0,                             offsetof(LinkConfig, features[NET_DEV_FEAT_RXALL])},
#line 43 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.Description",                          config_parse_string,                   0,                             offsetof(LinkConfig, description)},
#line 149 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"SR-IOV.MACAddress",                         config_parse_sr_iov_mac,               0,                             offsetof(LinkConfig, sr_iov_by_section)},
#line 115 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.TxCoalesceSec",                        config_parse_coalesce_sec,             0,                             offsetof(LinkConfig, coalesce.tx_coalesce_usecs)},
      {(char*)0},
#line 53 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.AlternativeName",                      config_parse_ifnames,                  IFNAME_VALID_ALTERNATIVE,      offsetof(LinkConfig, alternative_names)},
#line 137 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.IRQAffinity",                          config_parse_cpu_set,                  0,                             offsetof(LinkConfig, irq_affinity_cpus)},
#line 104 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.RxJumboBufferSize",                    config_parse_ring_buffer_or_channel,   0,                             offsetof(LinkConfig, ring.rx_jumbo)},
#line 39 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Match.Version",                             config_parse_net_condition,            CONDITION_VERSION,             offsetof(LinkConfig, conditions)},
      {(char*)0},
#line 51 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.NamePolicy",                           config_parse_name_policy,              0,                             offsetof(LinkConfig, name_policy)},
      {(char*)0}, {(char*)0},
#line 142 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"SR-IOV.VLANId",                             config_parse_sr_iov_uint32,            0,                             offsetof(LinkConfig, sr_iov_by_section)},
      {(char*)0},
#line 36 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Match.Virtualization",                      config_parse_net_condition,            CONDITION_VIRTUALIZATION,      offsetof(LinkConfig, conditions)},
#line 40 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Match.Credential",                          config_parse_net_condition,            CONDITION_CREDENTIAL,          offsetof(LinkConfig, conditions)},
      {(char*)0},
#line 63 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.BitsPerSecond",                        config_parse_si_uint64,                0,                             offsetof(LinkConfig, speed)},
      {(char*)0},
#line 79 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.TCPECNSegmentationOffload",            config_parse_tristate,                 0,                             offsetof(LinkConfig, features[NET_DEV_FEAT_TSO_ECN])},
#line 91 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.ReceiveVLANCTAGFilter",                config_parse_tristate,                 0,                             offsetof(LinkConfig, features[NET_DEV_FEAT_HW_VLAN_CTAG_FILTER])},
#line 54 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.AlternativeNamesPolicy",               config_parse_alternative_names_policy, 0,                             offsetof(LinkConfig, alternative_names_policy)},
#line 134 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.ReceivePacketSteeringCPUMask",         config_parse_rps_cpu_mask,             0,                             offsetof(LinkConfig, rps_cpu_mask)},
#line 129 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.RxMaxCoalescedHighFrames",             config_parse_coalesce_u32,             0,                             offsetof(LinkConfig, coalesce.rx_max_coalesced_frames_high)},
#line 30 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Match.Path",                                config_parse_match_strv,               0,                             offsetof(LinkConfig, match.path)},
      {(char*)0},
#line 93 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.NTupleFilter",                         config_parse_tristate,                 0,                             offsetof(LinkConfig, features[NET_DEV_FEAT_NTUPLE])},
#line 143 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"SR-IOV.QualityOfService",                   config_parse_sr_iov_uint32,            0,                             offsetof(LinkConfig, sr_iov_by_section)},
#line 131 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.TxMaxCoalescedHighFrames",             config_parse_coalesce_u32,             0,                             offsetof(LinkConfig, coalesce.tx_max_coalesced_frames_high)},
#line 65 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.AutoNegotiation",                      config_parse_tristate,                 0,                             offsetof(LinkConfig, autonegotiation)},
#line 92 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.TransmitVLANSTAGHardwareAcceleration", config_parse_tristate,                 0,                             offsetof(LinkConfig, features[NET_DEV_FEAT_HW_VLAN_STAG_TX])},
#line 135 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"/* IRQ affinity settings */"},
#line 81 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.TCP6SegmentationOffload",              config_parse_tristate,                 0,                             offsetof(LinkConfig, features[NET_DEV_FEAT_TSO6])},
#line 56 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.TransmitQueues",                       config_parse_rx_tx_queues,             0,                             offsetof(LinkConfig, txqueues)},
#line 28 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Match.PermanentMACAddress",                 config_parse_hw_addrs,                 0,                             offsetof(LinkConfig, match.permanent_hw_addr)},
#line 90 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.TransmitVLANCTAGHardwareAcceleration", config_parse_tristate,                 0,                             offsetof(LinkConfig, features[NET_DEV_FEAT_HW_VLAN_CTAG_TX])},
      {(char*)0}, {(char*)0}, {(char*)0}, {(char*)0},
#line 64 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.Duplex",                               config_parse_duplex,                   0,                             offsetof(LinkConfig, duplex)},
      {(char*)0},
#line 47 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.UnsetProperty",                        config_parse_udev_property_name,       0,                             offsetof(LinkConfig, unset_properties)},
#line 38 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Match.KernelVersion",                       config_parse_net_condition,            CONDITION_VERSION,             offsetof(LinkConfig, conditions)},
#line 146 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"SR-IOV.QueryReceiveSideScaling",            config_parse_sr_iov_boolean,           0,                             offsetof(LinkConfig, sr_iov_by_section)},
      {(char*)0}, {(char*)0},
#line 114 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.RxMaxCoalescedIrqFrames",              config_parse_coalesce_u32,             0,                             offsetof(LinkConfig, coalesce.rx_max_coalesced_frames_irq)},
#line 67 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.Port",                                 config_parse_port,                     0,                             offsetof(LinkConfig, port)},
#line 132 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.CoalescePacketRateSampleIntervalSec",  config_parse_coalesce_sec,             0,                             offsetof(LinkConfig, coalesce.rate_sample_interval)},
#line 109 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.AutoNegotiationFlowControl",           config_parse_tristate,                 0,                             offsetof(LinkConfig, autoneg_flow_control)},
      {(char*)0},
#line 118 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.TxMaxCoalescedIrqFrames",              config_parse_coalesce_u32,             0,                             offsetof(LinkConfig, coalesce.tx_max_coalesced_frames_irq)},
      {(char*)0},
#line 89 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.ReceiveVLANCTAGHardwareAcceleration",  config_parse_tristate,                 0,                             offsetof(LinkConfig, features[NET_DEV_FEAT_HW_VLAN_CTAG_RX])},
#line 100 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.CombinedChannels",                     config_parse_ring_buffer_or_channel,   0,                             offsetof(LinkConfig, channels.combined)},
      {(char*)0},
#line 107 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.RxFlowControl",                        config_parse_tristate,                 0,                             offsetof(LinkConfig, rx_flow_control)},
#line 60 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.GenericSegmentOffloadMaxBytes",        config_parse_iec_size,                 0,                             offsetof(LinkConfig, gso_max_size)},
#line 86 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.GenericReceiveOffloadList",            config_parse_tristate,                 0,                             offsetof(LinkConfig, features[NET_DEV_FEAT_GRO_FRAGLIST])},
      {(char*)0},
#line 61 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.GenericSegmentOffloadMaxSegments",     config_parse_uint32,                   0,                             offsetof(LinkConfig, gso_max_segments)},
#line 108 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.TxFlowControl",                        config_parse_tristate,                 0,                             offsetof(LinkConfig, tx_flow_control)},
#line 44 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"/* udev property */"},
#line 151 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"EnergyEfficientEthernet.Enable",            config_parse_tristate,                 0,                             offsetof(LinkConfig, eee_enabled)},
#line 77 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.GenericSegmentationOffload",           config_parse_tristate,                 0,                             offsetof(LinkConfig, features[NET_DEV_FEAT_GSO])},
#line 154 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"EnergyEfficientEthernet.LinkMode",          config_parse_advertise,                0,                             offsetof(LinkConfig, eee_advertise)},
#line 45 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.Property",                             config_parse_udev_property,            0,                             offsetof(LinkConfig, properties)},
#line 85 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.GenericReceiveOffloadHardware",        config_parse_tristate,                 0,                             offsetof(LinkConfig, features[NET_DEV_FEAT_GRO_HW])},
#line 138 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.IRQAffinityNUMA",                      config_parse_irq_affinity_numa,        0,                             offsetof(LinkConfig, irq_affinity_numa)},
#line 84 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.GenericReceiveOffload",                config_parse_tristate,                 0,                             offsetof(LinkConfig, features[NET_DEV_FEAT_GRO])},
#line 72 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"/* ethtool features */"},
#line 152 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"EnergyEfficientEthernet.TxLowPowerIdle",    config_parse_tristate,                 0,                             offsetof(LinkConfig, eee_tx_lpi_enabled)},
#line 87 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.GenericReceiveOffloadUDPForwarding",   config_parse_tristate,                 0,                             offsetof(LinkConfig, features[NET_DEV_FEAT_GRO_UDP_FWD])},
#line 106 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"/* ethtool pause parameters */"},
#line 119 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.StatisticsBlockCoalesceSec",           config_parse_coalesce_sec,             0,                             offsetof(LinkConfig, coalesce.stats_block_coalesce_usecs)},
#line 140 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.SR-IOVVirtualFunctions",               config_parse_sr_iov_num_vfs,           0,                             offsetof(LinkConfig, sr_iov_num_vfs)},
#line 99 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.OtherChannels",                        config_parse_ring_buffer_or_channel,   0,                             offsetof(LinkConfig, channels.other)},
#line 101 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"/* ethtool ring parameters */"},
#line 112 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.RxMaxCoalescedFrames",                 config_parse_coalesce_u32,             0,                             offsetof(LinkConfig, coalesce.rx_max_coalesced_frames)},
#line 153 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"EnergyEfficientEthernet.TxLowPowerIdleSec", config_parse_sec,                      0,                             offsetof(LinkConfig, eee_tx_lpi_timer_usec)},
#line 141 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"SR-IOV.VirtualFunction",                    config_parse_sr_iov_uint32,            0,                             offsetof(LinkConfig, sr_iov_by_section)},
#line 29 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Match.OriginalName",                        config_parse_match_ifnames,            0,                             offsetof(LinkConfig, match.ifname)},
      {(char*)0},
#line 116 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.TxMaxCoalescedFrames",                 config_parse_coalesce_u32,             0,                             offsetof(LinkConfig, coalesce.tx_max_coalesced_frames)},
#line 123 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.RxCoalesceLowSec",                     config_parse_coalesce_sec,             0,                             offsetof(LinkConfig, coalesce.rx_coalesce_usecs_low)},
      {(char*)0},
#line 74 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.TransmitChecksumOffload",              config_parse_tristate,                 0,                             offsetof(LinkConfig, features[NET_DEV_FEAT_TXCSUM])},
#line 88 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.LargeReceiveOffload",                  config_parse_tristate,                 0,                             offsetof(LinkConfig, features[NET_DEV_FEAT_LRO])},
#line 50 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.MACAddress",                           config_parse_hw_addr,                  0,                             offsetof(LinkConfig, hw_addr)},
#line 125 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.TxCoalesceLowSec",                     config_parse_coalesce_sec,             0,                             offsetof(LinkConfig, coalesce.tx_coalesce_usecs_low)},
#line 127 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.CoalescePacketRateHigh",               config_parse_coalesce_u32,             0,                             offsetof(LinkConfig, coalesce.pkt_rate_high)},
#line 59 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.MTUBytes",                             config_parse_mtu,                      AF_UNSPEC,                     offsetof(LinkConfig, mtu)},
      {(char*)0}, {(char*)0},
#line 122 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.CoalescePacketRateLow",                config_parse_coalesce_u32,             0,                             offsetof(LinkConfig, coalesce.pkt_rate_low)},
      {(char*)0}, {(char*)0},
#line 46 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.ImportProperty",                       config_parse_udev_property_name,       0,                             offsetof(LinkConfig, import_properties)},
#line 80 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.TCPMangleIdSegmentationOffload",       config_parse_tristate,                 0,                             offsetof(LinkConfig, features[NET_DEV_FEAT_TSO_MANGLEID])},
#line 110 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"/* ethtool coalesce settings */"},
#line 78 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.TCPSegmentationOffload",               config_parse_tristate,                 0,                             offsetof(LinkConfig, features[NET_DEV_FEAT_TSO])},
#line 68 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.MDI",                                  config_parse_mdi,                      0,                             offsetof(LinkConfig, mdi)},
#line 58 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.TransmitQueueLength",                  config_parse_txqueuelen,               0,                             offsetof(LinkConfig, txqueuelen)},
      {(char*)0},
#line 150 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"/* ethtool EEE settings */"},
#line 62 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"/* ethtool link settings */"},
#line 124 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.RxMaxCoalescedLowFrames",              config_parse_coalesce_u32,             0,                             offsetof(LinkConfig, coalesce.rx_max_coalesced_frames_low)},
#line 70 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.WakeOnLan",                            config_parse_wol,                      0,                             offsetof(LinkConfig, wol)},
#line 145 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"SR-IOV.MACSpoofCheck",                      config_parse_sr_iov_boolean,           0,                             offsetof(LinkConfig, sr_iov_by_section)},
#line 69 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"/* ethtool WoL settings */"},
#line 96 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"/* ethtool channels */"},
#line 126 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.TxMaxCoalescedLowFrames",              config_parse_coalesce_u32,             0,                             offsetof(LinkConfig, coalesce.tx_max_coalesced_frames_low)},
      {(char*)0}, {(char*)0},
#line 120 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.UseAdaptiveRxCoalesce",                config_parse_tristate,                 0,                             offsetof(LinkConfig, coalesce.use_adaptive_rx_coalesce)},
#line 73 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.ReceiveChecksumOffload",               config_parse_tristate,                 0,                             offsetof(LinkConfig, features[NET_DEV_FEAT_RXCSUM])},
#line 37 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Match.KernelCommandLine",                   config_parse_net_condition,            CONDITION_KERNEL_COMMAND_LINE, offsetof(LinkConfig, conditions)},
      {(char*)0}, {(char*)0},
#line 121 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.UseAdaptiveTxCoalesce",                config_parse_tristate,                 0,                             offsetof(LinkConfig, coalesce.use_adaptive_tx_coalesce)},
      {(char*)0},
#line 82 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.PartialGenericSegmentationOffload",    config_parse_tristate,                 0,                             offsetof(LinkConfig, features[NET_DEV_FEAT_GSO_PARTIAL])},
      {(char*)0}, {(char*)0},
#line 139 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"/* SR-IOV settings */"},
#line 136 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.IRQAffinityPolicy",                    config_parse_irq_affinity_policy,      0,                             offsetof(LinkConfig, irq_affinity_policy)},
#line 48 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"/* rtnl setlink */"},
      {(char*)0}, {(char*)0}, {(char*)0}, {(char*)0},
      {(char*)0},
#line 144 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"SR-IOV.VLANProtocol",                       config_parse_sr_iov_vlan_proto,        0,                             offsetof(LinkConfig, sr_iov_by_section)},
      {(char*)0}, {(char*)0}, {(char*)0}, {(char*)0},
      {(char*)0}, {(char*)0}, {(char*)0},
#line 83 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.UDPSegmentationOffload",               config_parse_warn_compat,              DISABLED_LEGACY,               0},
      {(char*)0}, {(char*)0}, {(char*)0}, {(char*)0},
#line 71 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.WakeOnLanPassword",                    config_parse_wol_password,             0,                             0},
      {(char*)0}, {(char*)0}, {(char*)0}, {(char*)0},
      {(char*)0}, {(char*)0}, {(char*)0}, {(char*)0},
      {(char*)0}, {(char*)0}, {(char*)0}, {(char*)0},
      {(char*)0}, {(char*)0}, {(char*)0}, {(char*)0},
      {(char*)0}, {(char*)0}, {(char*)0}, {(char*)0},
      {(char*)0}, {(char*)0}, {(char*)0},
#line 49 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"Link.MACAddressPolicy",                     config_parse_mac_address_policy,       0,                             offsetof(LinkConfig, mac_address_policy)},
      {(char*)0}, {(char*)0}, {(char*)0}, {(char*)0},
      {(char*)0}, {(char*)0}, {(char*)0}, {(char*)0},
      {(char*)0}, {(char*)0}, {(char*)0}, {(char*)0},
      {(char*)0}, {(char*)0},
#line 133 "../systemd/src/udev/net/link-config-gperf.gperf"
      {"/* Rx RPS CPU mask */"}
    };
#if (defined __GNUC__ && __GNUC__ + (__GNUC_MINOR__ >= 6) > 4) || (defined __clang__ && __clang_major__ >= 3)
#pragma GCC diagnostic pop
#endif

  if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH)
    {
      register unsigned int key = link_config_gperf_hash (str, len);

      if (key <= MAX_HASH_VALUE)
        {
          register const char *s = wordlist[key].section_and_lvalue;

          if (s && *str == *s && !strcmp (str + 1, s + 1))
            return &wordlist[key];
        }
    }
  return (struct ConfigPerfItem *) 0;
}
