static const char* const bpf_delegate_cmd_table[] = {
	"BPFMapCreate",
	"BPFMapLookupElem",
	"BPFMapUpdateElem",
	"BPFMapDeleteElem",
	"BPFMapGetNextKey",
	"BPFProgLoad",
	"BPFObjPin",
	"BPFObjGet",
	"BPFProgAttach",
	"BPFProgDetach",
	"BPFProgTestRun",
	"BPFProgGetNextId",
	"BPFMapGetNextId",
	"BPFProgGetFdById",
	"BPFMapGetFdById",
	"BPFObjGetInfoByFd",
	"BPFProgQuery",
	"BPFRawTracepointOpen",
	"BPFBtfLoad",
	"BPFBtfGetFdById",
	"BPFTaskFdQuery",
	"BPFMapLookupAndDeleteElem",
	"BPFMapFreeze",
	"BPFBtfGetNextId",
	"BPFMapLookupBatch",
	"BPFMapLookupAndDeleteBatch",
	"BPFMapUpdateBatch",
	"BPFMapDeleteBatch",
	"BPFLinkCreate",
	"BPFLinkUpdate",
	"BPFLinkGetFdById",
	"BPFLinkGetNextId",
	"BPFEnableStats",
	"BPFIterCreate",
	"BPFLinkDetach",
	"BPFProgBindMap",
	"BPFTokenCreate",
	"BPFProgStreamReadByFd",
	"BPFProgAssocStructOps",
};
static const char* const bpf_delegate_map_type_table[] = {
	"BPFMapTypeUnspec",
	"BPFMapTypeHash",
	"BPFMapTypeArray",
	"BPFMapTypeProgArray",
	"BPFMapTypePerfEventArray",
	"BPFMapTypePercpuHash",
	"BPFMapTypePercpuArray",
	"BPFMapTypeStackTrace",
	"BPFMapTypeCgroupArray",
	"BPFMapTypeLruHash",
	"BPFMapTypeLruPercpuHash",
	"BPFMapTypeLpmTrie",
	"BPFMapTypeArrayOfMaps",
	"BPFMapTypeHashOfMaps",
	"BPFMapTypeDevmap",
	"BPFMapTypeSockmap",
	"BPFMapTypeCpumap",
	"BPFMapTypeXskmap",
	"BPFMapTypeSockhash",
	"BPFMapTypeCgroupStorageDeprecated",
	"BPFMapTypeReuseportSockarray",
	"BPFMapTypePercpuCgroupStorageDeprecated",
	"BPFMapTypeQueue",
	"BPFMapTypeStack",
	"BPFMapTypeSkStorage",
	"BPFMapTypeDevmapHash",
	"BPFMapTypeStructOps",
	"BPFMapTypeRingbuf",
	"BPFMapTypeInodeStorage",
	"BPFMapTypeTaskStorage",
	"BPFMapTypeBloomFilter",
	"BPFMapTypeUserRingbuf",
	"BPFMapTypeCgrpStorage",
	"BPFMapTypeArena",
	"BPFMapTypeInsnArray",
};
static const char* const bpf_delegate_prog_type_table[] = {
	"BPFProgTypeUnspec",
	"BPFProgTypeSocketFilter",
	"BPFProgTypeKprobe",
	"BPFProgTypeSchedCls",
	"BPFProgTypeSchedAct",
	"BPFProgTypeTracepoint",
	"BPFProgTypeXdp",
	"BPFProgTypePerfEvent",
	"BPFProgTypeCgroupSkb",
	"BPFProgTypeCgroupSock",
	"BPFProgTypeLwtIn",
	"BPFProgTypeLwtOut",
	"BPFProgTypeLwtXmit",
	"BPFProgTypeSockOps",
	"BPFProgTypeSkSkb",
	"BPFProgTypeCgroupDevice",
	"BPFProgTypeSkMsg",
	"BPFProgTypeRawTracepoint",
	"BPFProgTypeCgroupSockAddr",
	"BPFProgTypeLwtSeg6local",
	"BPFProgTypeLircMode2",
	"BPFProgTypeSkReuseport",
	"BPFProgTypeFlowDissector",
	"BPFProgTypeCgroupSysctl",
	"BPFProgTypeRawTracepointWritable",
	"BPFProgTypeCgroupSockopt",
	"BPFProgTypeTracing",
	"BPFProgTypeStructOps",
	"BPFProgTypeExt",
	"BPFProgTypeLsm",
	"BPFProgTypeSkLookup",
	"BPFProgTypeNetfilter",
};
static const char* const bpf_delegate_attach_type_table[] = {
	"BPFCgroupInetIngress",
	"BPFCgroupInetEgress",
	"BPFCgroupInetSockCreate",
	"BPFCgroupSockOps",
	"BPFSkSkbStreamParser",
	"BPFSkSkbStreamVerdict",
	"BPFCgroupDevice",
	"BPFSkMsgVerdict",
	"BPFCgroupInet4Bind",
	"BPFCgroupInet6Bind",
	"BPFCgroupInet4Connect",
	"BPFCgroupInet6Connect",
	"BPFCgroupInet4PostBind",
	"BPFCgroupInet6PostBind",
	"BPFCgroupUdp4Sendmsg",
	"BPFCgroupUdp6Sendmsg",
	"BPFLircMode2",
	"BPFFlowDissector",
	"BPFCgroupSysctl",
	"BPFCgroupUdp4Recvmsg",
	"BPFCgroupUdp6Recvmsg",
	"BPFCgroupGetsockopt",
	"BPFCgroupSetsockopt",
	"BPFTraceRawTp",
	"BPFTraceFentry",
	"BPFTraceFexit",
	"BPFModifyReturn",
	"BPFLsmMac",
	"BPFTraceIter",
	"BPFCgroupInet4Getpeername",
	"BPFCgroupInet6Getpeername",
	"BPFCgroupInet4Getsockname",
	"BPFCgroupInet6Getsockname",
	"BPFXdpDevmap",
	"BPFCgroupInetSockRelease",
	"BPFXdpCpumap",
	"BPFSkLookup",
	"BPFXdp",
	"BPFSkSkbVerdict",
	"BPFSkReuseportSelect",
	"BPFSkReuseportSelectOrMigrate",
	"BPFPerfEvent",
	"BPFTraceKprobeMulti",
	"BPFLsmCgroup",
	"BPFStructOps",
	"BPFNetfilter",
	"BPFTcxIngress",
	"BPFTcxEgress",
	"BPFTraceUprobeMulti",
	"BPFCgroupUnixConnect",
	"BPFCgroupUnixSendmsg",
	"BPFCgroupUnixRecvmsg",
	"BPFCgroupUnixGetpeername",
	"BPFCgroupUnixGetsockname",
	"BPFNetkitPrimary",
	"BPFNetkitPeer",
	"BPFTraceKprobeSession",
	"BPFTraceUprobeSession",
	"BPFTraceFsession",
};
