const char* statx_attribute_to_name(uint64_t attr) {
        switch (attr) {
        case STATX_ATTR_AUTOMOUNT: return "STATX_ATTR_AUTOMOUNT";
        case STATX_ATTR_VERITY: return "STATX_ATTR_VERITY";
        case STATX_ATTR_IMMUTABLE: return "STATX_ATTR_IMMUTABLE";
        case STATX_ATTR_ENCRYPTED: return "STATX_ATTR_ENCRYPTED";
        case STATX_ATTR_APPEND: return "STATX_ATTR_APPEND";
        case STATX_ATTR_WRITE_ATOMIC: return "STATX_ATTR_WRITE_ATOMIC";
        case STATX_ATTR_COMPRESSED: return "STATX_ATTR_COMPRESSED";
        case STATX_ATTR_MOUNT_ROOT: return "STATX_ATTR_MOUNT_ROOT";
        case STATX_ATTR_DAX: return "STATX_ATTR_DAX";
        case STATX_ATTR_NODUMP: return "STATX_ATTR_NODUMP";
        default: return NULL;
        }
}
