/* SPDX-License-Identifier: LGPL-2.1-or-later */
#pragma once

#include "sd-varlink-idl.h"

extern const sd_varlink_symbol vl_type_DNSProtocol;
extern const sd_varlink_symbol vl_type_DNSOverTLSMode;
extern const sd_varlink_symbol vl_type_ResolveSupport;
extern const sd_varlink_symbol vl_type_ResolvConfMode;
extern const sd_varlink_symbol vl_type_ResourceKey;
extern const sd_varlink_symbol vl_type_ResourceRecord;
extern const sd_varlink_symbol vl_type_DNSServer;
extern const sd_varlink_symbol vl_type_DNSScope;
extern const sd_varlink_symbol vl_type_SearchDomain;
extern const sd_varlink_symbol vl_type_DNSConfiguration;

extern const sd_varlink_interface vl_interface_io_systemd_Resolve;
