/* ANSI-C code produced by gperf version 3.3 */
/* Command-line: gperf --output-file src/journal/journald-gperf.c ../systemd/src/journal/journald-gperf.gperf  */
/* Computed positions: -k'18,$' */

#if !((' ' == 32) && ('!' == 33) && ('"' == 34) && ('#' == 35) \
      && ('%' == 37) && ('&' == 38) && ('\'' == 39) && ('(' == 40) \
      && (')' == 41) && ('*' == 42) && ('+' == 43) && (',' == 44) \
      && ('-' == 45) && ('.' == 46) && ('/' == 47) && ('0' == 48) \
      && ('1' == 49) && ('2' == 50) && ('3' == 51) && ('4' == 52) \
      && ('5' == 53) && ('6' == 54) && ('7' == 55) && ('8' == 56) \
      && ('9' == 57) && (':' == 58) && (';' == 59) && ('<' == 60) \
      && ('=' == 61) && ('>' == 62) && ('?' == 63) && ('A' == 65) \
      && ('B' == 66) && ('C' == 67) && ('D' == 68) && ('E' == 69) \
      && ('F' == 70) && ('G' == 71) && ('H' == 72) && ('I' == 73) \
      && ('J' == 74) && ('K' == 75) && ('L' == 76) && ('M' == 77) \
      && ('N' == 78) && ('O' == 79) && ('P' == 80) && ('Q' == 81) \
      && ('R' == 82) && ('S' == 83) && ('T' == 84) && ('U' == 85) \
      && ('V' == 86) && ('W' == 87) && ('X' == 88) && ('Y' == 89) \
      && ('Z' == 90) && ('[' == 91) && ('\\' == 92) && (']' == 93) \
      && ('^' == 94) && ('_' == 95) && ('a' == 97) && ('b' == 98) \
      && ('c' == 99) && ('d' == 100) && ('e' == 101) && ('f' == 102) \
      && ('g' == 103) && ('h' == 104) && ('i' == 105) && ('j' == 106) \
      && ('k' == 107) && ('l' == 108) && ('m' == 109) && ('n' == 110) \
      && ('o' == 111) && ('p' == 112) && ('q' == 113) && ('r' == 114) \
      && ('s' == 115) && ('t' == 116) && ('u' == 117) && ('v' == 118) \
      && ('w' == 119) && ('x' == 120) && ('y' == 121) && ('z' == 122) \
      && ('{' == 123) && ('|' == 124) && ('}' == 125) && ('~' == 126))
/* The character set is not based on ISO-646.  */
#error "gperf generated tables don't work with this execution character set. Please report a bug to <bug-gperf@gnu.org>."
#endif

#line 2 "../systemd/src/journal/journald-gperf.gperf"

_Pragma("GCC diagnostic ignored \"-Wimplicit-fallthrough\"")
#if __GNUC__ >= 15
_Pragma("GCC diagnostic ignored \"-Wzero-as-null-pointer-constant\"")
#endif
#include <stddef.h>
#include "conf-parser.h"
#include "journald-config.h"
#include <string.h>

#define TOTAL_KEYWORDS 33
#define MIN_WORD_LENGTH 12
#define MAX_WORD_LENGTH 28
#define MIN_HASH_VALUE 12
#define MAX_HASH_VALUE 63
/* maximum key range = 52, duplicates = 0 */

#ifdef __GNUC__
__inline
#else
#ifdef __cplusplus
inline
#endif
#endif
static unsigned int
journald_gperf_hash (register const char *str, register size_t len)
{
  static const unsigned char asso_values[] =
    {
      64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
      64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
      64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
      64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
      64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
      64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
      64, 64, 64, 64, 64, 64, 15,  0, 64, 64,
       0, 64, 64, 20, 64, 25, 64, 30, 64, 64,
      64, 64, 64, 20, 64, 15, 64, 20, 64, 64,
      64, 64, 64, 64, 64, 64, 64, 15, 64,  0,
      64,  5, 64,  0, 10, 40, 64, 64,  0, 64,
      64, 10, 10, 64, 64,  5, 10, 64, 35, 64,
       0, 10, 64, 64, 64, 64, 64, 64, 64, 64,
      64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
      64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
      64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
      64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
      64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
      64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
      64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
      64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
      64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
      64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
      64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
      64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
      64, 64, 64, 64, 64, 64
    };
  register unsigned int hval = len;

  switch (hval)
    {
      default:
        hval += asso_values[(unsigned char)str[17]];
#if (defined __cplusplus && (__cplusplus >= 201703L || (__cplusplus >= 201103L && defined __clang__ && __clang_major__ + (__clang_minor__ >= 9) > 3))) || (defined __STDC_VERSION__ && __STDC_VERSION__ >= 202000L && ((defined __GNUC__ && __GNUC__ >= 10) || (defined __clang__ && __clang_major__ >= 9)))
      [[fallthrough]];
#elif (defined __GNUC__ && __GNUC__ >= 7) || (defined __clang__ && __clang_major__ >= 10)
      __attribute__ ((__fallthrough__));
#endif
      /*FALLTHROUGH*/
      case 17:
      case 16:
      case 15:
      case 14:
      case 13:
      case 12:
        break;
    }
  return hval + asso_values[(unsigned char)str[len - 1]];
}

const struct ConfigPerfItem *
journald_gperf_lookup (register const char *str, register size_t len)
{
#if (defined __GNUC__ && __GNUC__ + (__GNUC_MINOR__ >= 6) > 4) || (defined __clang__ && __clang_major__ >= 3)
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wmissing-field-initializers"
#endif
  static const struct ConfigPerfItem wordlist[] =
    {
      {(char*)0}, {(char*)0}, {(char*)0}, {(char*)0},
      {(char*)0}, {(char*)0}, {(char*)0}, {(char*)0},
      {(char*)0}, {(char*)0}, {(char*)0}, {(char*)0},
#line 24 "../systemd/src/journal/journald-gperf.gperf"
      {"Journal.Seal",                 config_parse_tristate,          0, offsetof(JournalConfig, seal)},
      {(char*)0}, {(char*)0},
#line 55 "../systemd/src/journal/journald-gperf.gperf"
      {"Journal.LineMax",              config_parse_line_max,          0, offsetof(JournalConfig, line_max)},
#line 25 "../systemd/src/journal/journald-gperf.gperf"
      {"Journal.ReadKMsg",             config_parse_tristate,          0, offsetof(JournalConfig, read_kmsg)},
      {(char*)0},
#line 41 "../systemd/src/journal/journald-gperf.gperf"
      {"Journal.MaxFileSec",           config_parse_sec,               0, offsetof(JournalConfig, max_file_usec)},
      {(char*)0},
#line 22 "../systemd/src/journal/journald-gperf.gperf"
      {"Journal.Storage",              config_parse_storage,           0, offsetof(JournalConfig, storage)},
#line 23 "../systemd/src/journal/journald-gperf.gperf"
      {"Journal.Compress",             config_parse_compress,          0, offsetof(JournalConfig, compress)},
#line 54 "../systemd/src/journal/journald-gperf.gperf"
      {"Journal.SplitMode",            config_parse_split_mode,        0, offsetof(JournalConfig, split_mode)},
#line 26 "../systemd/src/journal/journald-gperf.gperf"
      {"Journal.Audit",                config_parse_audit_set_mode,    0, offsetof(JournalConfig, set_audit)},
      {(char*)0},
#line 47 "../systemd/src/journal/journald-gperf.gperf"
      {"Journal.TTYPath",              config_parse_path,              0, offsetof(JournalConfig, tty_path)},
#line 36 "../systemd/src/journal/journald-gperf.gperf"
      {"Journal.RuntimeMaxUse",        config_parse_iec_uint64,        0, offsetof(JournalConfig, runtime_storage_metrics.max_use)},
#line 35 "../systemd/src/journal/journald-gperf.gperf"
      {"Journal.SystemMaxFiles",       config_parse_uint64,            0, offsetof(JournalConfig, system_storage_metrics.n_max_files)},
#line 39 "../systemd/src/journal/journald-gperf.gperf"
      {"Journal.RuntimeMaxFiles",      config_parse_uint64,            0, offsetof(JournalConfig, runtime_storage_metrics.n_max_files)},
#line 44 "../systemd/src/journal/journald-gperf.gperf"
      {"Journal.ForwardToConsole",     config_parse_tristate,          0, offsetof(JournalConfig, forward_to_console)},
#line 33 "../systemd/src/journal/journald-gperf.gperf"
      {"Journal.SystemMaxFileSize",    config_parse_iec_uint64,        0, offsetof(JournalConfig, system_storage_metrics.max_size)},
#line 37 "../systemd/src/journal/journald-gperf.gperf"
      {"Journal.RuntimeMaxFileSize",   config_parse_iec_uint64,        0, offsetof(JournalConfig, runtime_storage_metrics.max_size)},
#line 49 "../systemd/src/journal/journald-gperf.gperf"
      {"Journal.MaxLevelSyslog",       config_parse_log_level,         0, offsetof(JournalConfig, max_level_syslog)},
#line 38 "../systemd/src/journal/journald-gperf.gperf"
      {"Journal.RuntimeKeepFree",      config_parse_iec_uint64,        0, offsetof(JournalConfig, runtime_storage_metrics.keep_free)},
      {(char*)0},
#line 52 "../systemd/src/journal/journald-gperf.gperf"
      {"Journal.MaxLevelWall",         config_parse_log_level,         0, offsetof(JournalConfig, max_level_wall)},
#line 48 "../systemd/src/journal/journald-gperf.gperf"
      {"Journal.MaxLevelStore",        config_parse_log_level,         0, offsetof(JournalConfig, max_level_store)},
#line 34 "../systemd/src/journal/journald-gperf.gperf"
      {"Journal.SystemKeepFree",       config_parse_iec_uint64,        0, offsetof(JournalConfig, system_storage_metrics.keep_free)},
#line 51 "../systemd/src/journal/journald-gperf.gperf"
      {"Journal.MaxLevelConsole",      config_parse_log_level,         0, offsetof(JournalConfig, max_level_console)},
      {(char*)0},
#line 32 "../systemd/src/journal/journald-gperf.gperf"
      {"Journal.SystemMaxUse",         config_parse_iec_uint64,        0, offsetof(JournalConfig, system_storage_metrics.max_use)},
#line 45 "../systemd/src/journal/journald-gperf.gperf"
      {"Journal.ForwardToWall",        config_parse_tristate,          0, offsetof(JournalConfig, forward_to_wall)},
#line 53 "../systemd/src/journal/journald-gperf.gperf"
      {"Journal.MaxLevelSocket",       config_parse_log_level,         0, offsetof(JournalConfig, max_level_socket)},
#line 42 "../systemd/src/journal/journald-gperf.gperf"
      {"Journal.ForwardToSyslog",      config_parse_tristate,          0, offsetof(JournalConfig, forward_to_syslog)},
      {(char*)0},
#line 29 "../systemd/src/journal/journald-gperf.gperf"
      {"Journal.RateLimitInterval",    config_parse_sec,               0, offsetof(JournalConfig, ratelimit_interval)},
#line 43 "../systemd/src/journal/journald-gperf.gperf"
      {"Journal.ForwardToKMsg",        config_parse_tristate,          0, offsetof(JournalConfig, forward_to_kmsg)},
#line 31 "../systemd/src/journal/journald-gperf.gperf"
      {"Journal.RateLimitBurst",       config_parse_unsigned,          0, offsetof(JournalConfig, ratelimit_burst)},
#line 30 "../systemd/src/journal/journald-gperf.gperf"
      {"Journal.RateLimitIntervalSec", config_parse_sec,               0, offsetof(JournalConfig, ratelimit_interval)},
      {(char*)0},
#line 50 "../systemd/src/journal/journald-gperf.gperf"
      {"Journal.MaxLevelKMsg",         config_parse_log_level,         0, offsetof(JournalConfig, max_level_kmsg)},
      {(char*)0}, {(char*)0},
#line 46 "../systemd/src/journal/journald-gperf.gperf"
      {"Journal.ForwardToSocket",      config_parse_forward_to_socket, 0, offsetof(JournalConfig, forward_to_socket)},
      {(char*)0}, {(char*)0}, {(char*)0}, {(char*)0},
#line 27 "../systemd/src/journal/journald-gperf.gperf"
      {"Journal.SyncIntervalSec",      config_parse_sec,               0, offsetof(JournalConfig, sync_interval_usec)},
      {(char*)0}, {(char*)0}, {(char*)0}, {(char*)0},
#line 40 "../systemd/src/journal/journald-gperf.gperf"
      {"Journal.MaxRetentionSec",      config_parse_sec,               0, offsetof(JournalConfig, max_retention_usec)}
    };
#if (defined __GNUC__ && __GNUC__ + (__GNUC_MINOR__ >= 6) > 4) || (defined __clang__ && __clang_major__ >= 3)
#pragma GCC diagnostic pop
#endif

  if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH)
    {
      register unsigned int key = journald_gperf_hash (str, len);

      if (key <= MAX_HASH_VALUE)
        {
          register const char *s = wordlist[key].section_and_lvalue;

          if (s && *str == *s && !strcmp (str + 1, s + 1))
            return &wordlist[key];
        }
    }
  return (struct ConfigPerfItem *) 0;
}
